/* GSI - general sound interface   [-*- C++ -*-]
 *
 * mixer.c - shared sound server, mixer setting/reading
 *   - sept 1996/may 1997/dec 1997/march-may 1998
 *
 * Copyright (C) 1996,1997,1998 W.H.Scholten
 *
 * Lower level API dependencies:
 *   - DSP mode setting: OSS but should not be too difficult to adapt
 *     to other systems.
 *   - OSS include locations are Linux/FreeBSD specific.
 *     Locations for other OS's?
 *   - sequencer/synth (MIDI's) playing: very OSS specific.
 */

/* This software is licensed under the CGPL version 0.4 or later
 * (see the file LICENSE.CGPL for details)
 */




/*__________________INCLUDES___________________*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <errno.h>

#include <gsi/config.h>
#ifdef USE_OSS
#ifdef __linux__
#  include <sys/soundcard.h>
#endif
#ifdef __FreeBSD__
#  include <machine/soundcard.h>
#  include <sys/errno.h>
#endif
#endif
#ifdef USE_NEW_BSD
#  error code for new bsd driver not implemented
#endif
#ifdef USE_DEV_AUDIO
#  warning some includes for /dev/audio may be missing
#endif

#include <gsi/mytypes.h>
#include <gsi/debug.h>
#define _SNDSERVER_
#include <gsi/gsi_commands.h>
#include <gsi/gsi_server_version.h>
#include "midi.h"

/*__________________DEFINITIES___________________*/

#define min(x,y) ((x)>(y) ? (y): (x))
#define max(x,y) ((x)<(y) ? (y): (x))


/*__________________EXTERN___________________*/
extern int  dsp;

int  GSI_master_vol=-1;
int  GSI_pcm_vol=-1;
int  GSI_synth_vol=-1;
int  GSI_cd_vol=-1;
int  mixer_ok=False;
int  emulate_master_volume=False;


/*__________________VARIABELEN___________________*/

/*__________________FUNCTIE PROTOTYPES___________________*/
#ifdef USE_OSS
static void set_oss_volume(int device, int vol);
static int  get_oss_volume(int device);
char *mixerdev="/dev/mixer";
int  mixer_devices;
#endif


void set_master_volume(int vol);
void set_pcm_volume(int vol);
void set_synth_volume(int vol);
void set_cd_volume(int vol);
int  get_master_volume();
int  get_pcm_volume();
int  get_synth_volume();
int  get_cd_volume();

/*__________________PROGRAMMA___________________*/


int set_mixer(int mixer_device, int vol) {
	switch(mixer_device) {
	case GSI_MASTER_VOLUME:
		set_master_volume(vol); break;
	case GSI_PCM_VOLUME:
		set_pcm_volume(vol); break;
	case GSI_SYNTH_VOLUME:
		set_synth_volume(vol); break;
	case GSI_CD_VOLUME:
		set_cd_volume(vol); break;

	}
	return 0;
}
int read_mixer(int mixer_device) {
	switch(mixer_device) {
	case GSI_MASTER_VOLUME:
		return get_master_volume(mixer_device); break;
	case GSI_PCM_VOLUME:
		return get_pcm_volume(mixer_device); break;
	case GSI_SYNTH_VOLUME:
		return get_synth_volume(mixer_device); break;
	case GSI_CD_VOLUME:
		return get_cd_volume(mixer_device); break;
	}
	return 0;
}


/*---------------------------------------------------------------------------*
 *	Set sound volumes
 *  Mixer volume settings:
 *    VOLUME : master volume, i.e. headphone and line out.
 *    PCM    : dsp volume
 *    SYNTH  : onboard midi playing device
 *
 *  saving volumes is done to remove any rounding problems with repeated
 *  reading/writing (device values read may not be the same as what has been
 *  set)
 *---------------------------------------------------------------------------*/
void set_master_volume(int vol) {
	GSI_master_vol=vol; 
	if (emulate_master_volume) {
		set_pcm_volume(GSI_pcm_vol);
		set_synth_volume(GSI_synth_vol);
		set_cd_volume(GSI_cd_vol);
	}
	else {
#ifdef USE_OSS
		set_oss_volume(SOUND_MIXER_WRITE_VOLUME, vol);
#endif
	}
}
void set_pcm_volume(int vol) {
	GSI_pcm_vol=vol;
	if (emulate_master_volume) vol=(vol*GSI_master_vol)>>8;
#ifdef USE_OSS
	set_oss_volume(SOUND_MIXER_WRITE_PCM, vol);
#endif
}
void set_synth_volume(int vol) {
	GSI_synth_vol=vol;
	if (emulate_master_volume) vol=(vol*GSI_master_vol)>>8;
#ifdef USE_OSS
	if (mixer_devices & SOUND_MASK_SYNTH)
		set_oss_volume(SOUND_MIXER_WRITE_SYNTH, vol);
#endif
}
void set_cd_volume(int vol) {
	GSI_cd_vol=vol;
	if (emulate_master_volume) vol=(vol*GSI_master_vol)>>8;
#ifdef USE_OSS
	set_oss_volume(SOUND_MIXER_WRITE_CD, vol);
#endif
}

int get_master_volume() {
	if (GSI_master_vol<0) {
#ifdef USE_OSS
		GSI_master_vol=get_oss_volume(SOUND_MIXER_READ_VOLUME);
#endif
	}
	return GSI_master_vol;
}
int get_pcm_volume() {
	if (GSI_pcm_vol<0) {
#ifdef USE_OSS
		GSI_pcm_vol=get_oss_volume(SOUND_MIXER_READ_PCM);
#endif
	}
	return GSI_pcm_vol;
}
int get_synth_volume() {
	if (GSI_synth_vol<0) {
#ifdef USE_OSS
		if (mixer_devices & SOUND_MASK_SYNTH)
			GSI_synth_vol=get_oss_volume(SOUND_MIXER_READ_SYNTH);
#endif
	}
	return GSI_synth_vol;
}
int get_cd_volume() {
	if (GSI_cd_vol<0) {
#ifdef USE_OSS
		GSI_cd_vol=get_oss_volume(SOUND_MIXER_READ_CD);
#endif
	}
	return GSI_cd_vol;
}


/*---------------------------------------------------------------------------*
 *	Initialise mixer, checks whether we have a master volume and reads the
 *	current volumes.
 *---------------------------------------------------------------------------*/
int init_mixer() {
	int  mixer;

	/* Init the mixer */
#ifdef USE_OSS
	mixer = open(mixerdev, O_WRONLY | O_NONBLOCK);
	if(mixer==-1) {
		mixer_ok=0;
		fprintf(stderr,"GSI: Couldn't open mixer %s\n", mixerdev);
	}
	else {
		ioctl(mixer, SOUND_MIXER_READ_DEVMASK, &mixer_devices);
		close(mixer);
	}
	if (!(mixer_devices & SOUND_MASK_VOLUME)) {
		emulate_master_volume=True;
		GSI_master_vol=256;
	}
	else get_master_volume();
#endif

	get_pcm_volume();
	get_synth_volume();
	get_cd_volume();

	return 0;
}


/*---------------------------------------------------------------------------*
 *	Get/set sound volumes with OSS
 *---------------------------------------------------------------------------*/
#ifdef USE_OSS
/* We expect volumes in the range 0-255 */
static void set_oss_volume(int device, int vol) {
	int mixer, volume;

	if (mixer_ok) {
		mixer=open(mixerdev, O_WRONLY | O_NONBLOCK);
		volume = (vol & 65535); //maybe use range 0-65535
		if (volume >255) volume=255;
		volume=(volume*100)/255; /* OSS uses a scale 0-100 */
		DEB1(fprintf(stderr, "GSI: set volume: sound_volume=%d -> %d real\n", volume, vol));
		volume |=volume<<8; /* Left and right channels */
		ioctl(mixer, device, &volume);
		close(mixer);
	}
}

/* We return volumes in the range 0-255 */
static int get_oss_volume(int device) {
	int mixer, volume;
	int volume_L, volume_R;

	if (mixer_ok) {
		mixer=open(mixerdev, O_RDONLY | O_NONBLOCK);
		ioctl(mixer, device, &volume);
		volume_L = volume &255;
		volume_R = (volume>>8) & 255;
//		volume =((volume_L*255)/100) | (((volume_L*255)/100)<<8);
		volume =((volume_L*255)/100); //We currently only return mono (L)
		DEB1(fprintf(stderr, "GSI: get_volume: sound_volume=%d -> %d real\n", volume, volume_L));
		close(mixer);
	}
	return volume &255; /* ignore R for now, should all HW volume settings be made stereo ? */
}
#endif /* USE_OSS */

/*---------------------------------------------------------------------------*
 *	EOF mixer.c
 *---------------------------------------------------------------------------*/
